 ###########################################################################
 # 11/5/2001: David V. Pynadath (pynadath@isi.edu)
 #
 # Agent: generic class for defining an agent and its capabilities
 #
 # Team: generic class for defining a team of agents
 #
 # ComMTDP: the COM-MTDP class
 # 
 ###########################################################################
import copy
import string
from types import *

from States import StateSpace
from States import EnumeratedSubspace

def strip(state):
    s = {}
    for key in state.keys():
        if key[0] != '_':
            s[key] = state[key]
    return s

class Agent:
    def __init__(self,actions, # A_i: a list
		 observations, # Omega_i: a list
		 messages,     # Sigma_i: a list
		 name='Generic Agent'): # a label string
	self.name = name
	# A_i
	self.actions = actions[:]
	# Omega_i
	self.observations = observations
	# Sigma_i
	self.messages = messages
	# B_i^0
	self.beliefs = {}

    def legalActions(self,state):
	"""Default method for specifying the set (i.e., list) of
	actions that this agent can choose from in the given state"""
	return self.actions[:]

    def legalMessages(self,state):
	"""Default method for specifying the set (i.e., list) of
	messages that this agent can choose from in the given state"""
	return self.messages[:]

    

class Team:
    def __init__(self,agents): # a list of Agent objects
	self.agents = {}
	if len(agents) == 0:
	    print 'Team::__init__: empty list of agents'
	    return
	for agent in agents:
	    self.agents[agent.name] = agent

    def composeActions(self,actions):
	"""Translates a dictionary of agent:action pairs into a single 
	string"""
	agentList = self.agents.keys()
	actionString = actions[agentList[0]]
	for agent in agentList[1:]:
	    actionString = actionString + ' ' + actions[agent]
	return actionString

    def decomposeActions(self,actionString):
	"""Translates a string into a dictionary of agent:action
	pairs"""
	actionList = string.split(actionString)
	actions = {}
	for agent in range(len(self.agents)):
	    actions[self.agents.keys()[agent]] = actionList[agent]
	return actions

    # This method should probably be in ComMTDP
    def individualObs(self,state,actions,observation,agent):
	"""Default method for returning the probability distribution
	over observations for an individual agent.  Returns a uniform
	distribution.  Should be overwritten."""
	print "Team: unimplemented individualObs method"
	if observation in self.agents[agent].observations:
	    return 1.0 / float(len(self.agents[agent].observations))
	else:
	    return 0.0

    # This method should probably be in ComMTDP
    def observableProb(self,state,actions,observation,agent):
	"""Handles the observation function for features that are
	completely observable.  If the observation does not match the
	true state on the completely observable features, then this
	function returns 0.0; otherwise, it returns 1.0"""
	for key in observation.keys():
	    if key[0] != '_' and \
	       self.agents[agent].observations[key] == 'observable':
		if state[key] != observation[key]:
		    return 0.0
	else:
	    return 1.0

class ComMTDP:
    version = 1.0
    def __init__(self,states,team,name='Generic COM-MTDP'):
	self.states = states
	self.team = team
	self.agents = []
	for agent in self.team.agents.keys():
	    self.agents.append(self.team.agents[agent])
	self.name = name

    # R
    def reward(self,state,actions,messages):
	return self.rewardAct(state,actions) + \
	       self.rewardCom(state,messages)

    # R_A
    def rewardAct(self,state,actions):
	print "ComMTDP: unimplemented domain-level reward method"
	return 0.0

    # R_Sigma
    def rewardCom(self,state,messages):
	print "ComMTDP: unimplemented coordination-level reward method"
	return 0.0

    # P
    def probability(self,orig,dest,actions):
	print "ComMTDP: unimplemented transition probability method"
	return 1.0 / float(self.states.size)

    def evaluatePolicy(self,domPolicy,comPolicy,state,horizon=100,debug=0):
	"""Computes expected reward of following the domain- and
	communication-level policies over the finite horizon
	specified, from the initial state specified"""
	# Create state of execution, which includes state of the
	# world, as well as belief states of individual agents
	stateList = [{'_world': state,'_prob':1.0,'_value':0.0,
                      '_actions':{},'_epoch':0}]
	for s in stateList:
	    for agent in self.agents:
		s[agent.name] = self.initialStateEstimator(agent)
	return self._evaluatePolicy(stateList,domPolicy,comPolicy,
                                    horizon,debug)
        
    def _evaluatePolicy(self,stateList,domPolicy,comPolicy,horizon,debug):
        results = {'Reward':0.0,
                   'Messages':0.0}
        if len(stateList) == 0:
            return results
	for epoch in range(stateList[0]['_epoch'],horizon):
	    if debug:
		print '========='
		print 'Epoch',epoch
		print '========='
            self.ProjectObservations(stateList,None,[],epoch,debug)
            results['Messages'] = results['Messages'] + \
                                  self.ExecuteCommunication(stateList,
                                                            comPolicy,debug)
            results['Reward'] = results['Reward'] \
                                + self.ExecuteActions(stateList,domPolicy,
                                                      debug)
            self.ProjectWorldDynamics(stateList)
	if debug:
	    print 'EValue:',results['Reward']
            print '# Msgs:',results['Messages']
	return results

    def _generateTeamMessages(self,state,agents,msgList):
        """Generates all possible messages over all agents"""
        if len(agents) == 0:
            return msgList
        else:
            # Generate set of possible messages for the current agent
            agent = agents[0]
            for teamMsg in msgList[:]:
                for msg in agent.legalMessages(state[agent.name]):
                    newMsg = copy.copy(teamMsg)
                    if msg:
                        newMsg[agent.name] = msg
                    msgList.append(newMsg)
                msgList.remove(teamMsg)
            agents.remove(agent)
            return self._generateTeamMessages(state,agents,msgList)
        
    def _generateTeamObservations(self,state,actions,agents,obsList,debug=0):
	"""Generates all possible observations over all agents"""
	if len(agents) == 0:
	    return obsList
	else:
	    # Find the current agent's observations
	    agent = agents[0]
	    agentObsList = [{'_type':'observation'}]
	    self._generateAgentObservations(state,actions,agent,self.states.features.keys()[:],agentObsList,debug)
	    # Compose this list with current list of possible observations
	    for obs in obsList[:]:
		obsList.remove(obs)
		for agentObs in agentObsList:
		    newObs = copy.copy(obs)
		    newObs[agent.name] = agentObs
		    newObs['_prob'] = newObs['_prob'] * \
				      self.team.individualObs(state,actions,agentObs,agent.name)
		    if newObs['_prob'] > 0.0:
			obsList.append(newObs)
	    agents.remove(agent)
	    return self._generateTeamObservations(state,actions,agents,obsList)

    def _generateAgentObservations(self,state,actions,agent,
				   features,observations,debug=0):
	"""Generates all possible observations for an individual agent"""
	if len(features) == 0:
	    if debug:
		print 'Agent',agent.name,'observes:',observations
	    return observations
	else:
	    feature = features[0]
	    if agent.observations.has_key(feature):
		if agent.observations[feature] == 'observable':
		    # For observable features, there is only one
		    # possible observation
		    for obs in observations:
			obs[feature] = state[feature]
		else:
		    # For partially observable features, we must
		    # consider all of the possible observations and
		    # their probabilities.
		    for obs in observations[:]:
			observations.remove(obs)
			for value in agent.observations[feature]:
			    newObs = copy.copy(obs)
			    newObs[feature] = value
			    observations.append(newObs)
	    else:
		# Feature is unobservable
		pass
	    features.remove(feature)
	    return self._generateAgentObservations(state,actions,
						   agent,features,
						   observations)			
    def ProjectObservations(self,states,agent=None,beliefs=[],
                            epoch=-1,debug=0):
        # Observation phase
        if debug:
            print
            print '--------'
            print 'Observation phase:'
        for s in states[:]:
            if debug:
                print '--------'
                print 'Examining state:',strip(s['_world'])
            # Draw observations of initial world state (under no actions)
            observations = [{'_prob':1.0}]
            self._generateTeamObservations(s['_world'],s['_actions'],
                                           self.agents[:],observations,
                                           debug)
            if debug:
                print 'Possible observations:'
                for obs in observations:
                    print '\t',strip(obs)
            # Eliminate observations inconsistent with state
            belief = None
            for belief in beliefs:
                if belief['_type'] == 'observation' and \
                   belief['_epoch'] == epoch:
                    totalProb = 0.0
                    break
            else:
                totalProb = 1.0
                belief = None
            if belief:
                for obs in observations[:]:
                    for feature in obs[agent].keys():
                        if feature[0] != '_' and \
                           obs[agent][feature] != belief[feature]:
                            break
                    else:
                        feature = None
                    if feature:
                        # Observation is inconsistent
                        observations.remove(obs)
                    else:
                        # Observation is consistent
                        totalProb = totalProb + obs['_prob']
            if len(observations) == 0:
                # The state itself is inconsistent with observations
                for state in states:
                    if not state is s:
                        state['_prob'] = state['_prob'] / (1.0 - s['_prob'])
            else:
                # Update belief state based on new observations
                if s.has_key('_parent'):
                    s['_children'] = []
                for obs in observations:
                    newState = copy.copy(s)
                    for a in self.agents:
                        newState[a.name] = self.preComStateEstimator(a,s[a.name],obs[a.name],epoch)
                    # Use normalized probability
                    newState['_prob'] = newState['_prob'] \
                                        * obs['_prob'] / totalProb
                    if s.has_key('_parent'):
                        newState['_parent'] = s
                        s['_children'].append(newState)
                    states.append(newState)
            states.remove(s)

    def ExecuteCommunication(self,states,comPolicy,debug=0):
        # Communication phase
        if debug:
            print
            print '--------'
            print 'Communication phase:'
        msgCount = 0.0
        for s in states[:]:
            if debug:
                print '--------'
                print 'Examining state:',strip(s)
            if comPolicy:
                # Execute each agent's communication policy to
                # generate set of messages
                messages = [{'_type':'message'}]
                for agent in self.agents:
                    contents = comPolicy[agent.name].execute(s[agent.name])
                    if not contents or type(contents) is StringType:
                        contents = [contents]
                    for msg in messages[:]:
                        for content in contents:
                            newMsg = copy.copy(msg)
                            if content:
                                newMsg[agent.name] = content
                                msgCount = msgCount + s['_prob']
                            messages.append(newMsg)
                        messages.remove(msg)
            else:
                # Generate all possible messages
		messages = self._generateTeamMessages(s,self.agents[:],[{}])
            if s.has_key('_parent'):
                s['_children'] = {}
            for msg in messages:
                newState = copy.copy(s)
                msg['_type'] = 'message'
                if debug:
                    print 'Messages:',strip(msg)
                # Update agents' beliefs based on messages exchanged
                for agent in self.agents:
                    newState[agent.name] = self.postComStateEstimator(agent,s[agent.name],msg,s['_epoch'])
                # Update value based on communication cost
                newState['_value'] = self.rewardCom(s['_world'],msg)
                newState['_messages'] = msg
                if s.has_key('_parent'):
                    newState['_parent'] = s
                    s['_children'][`msg`] = newState
                states.append(newState)
            states.remove(s)
        return msgCount

    def ExecuteActions(self,states,domPolicy,debug=0):
        # Action phase
        if debug:
            print
            print '--------'
            print 'Action phase:'
        value = 0.0
        for s in states[:]:
            if debug:
                print '--------'
                print 'Examining state:',strip(s['_world'])
                print 'Beliefs:',strip(s)
            # Execute each agent's domain-level policy to
            # generate set of actions
            s['_actions'] = {}
            for agent in self.agents:
                act = domPolicy[agent.name].execute(s[agent.name])
                if act:
                    s['_actions'][agent.name] = act
            if debug:
                print 'Actions:',s['_actions']
            # Update value based on action cost
            s['_value'] = s['_value'] \
                          + self.rewardAct(s['_world'],s['_actions'])
            # Update overall value
            value = value + s['_prob'] * s['_value']
            if debug:
                print 'Reward:',s['_value']
        return value

    def ProjectWorldDynamics(self,states,debug=0):
        # World Dynamics phase
        if debug:
            print
            print '--------'
            print 'World dynamics phase:'
        for s in states[:]:
            if debug:
                print '--------'
                print 'Examining state:',strip(s['_world'])
                print 'Beliefs:',strip(s)
            orig = self.states.state2index(s['_world'])
            action = self.team.composeActions(s['_actions'])
            if s.has_key('_parent'):
                s['_children'] = {}
            for dest in self.states.children[orig][action]:
                destState = self.states.index2state(dest)
                newState = copy.copy(s)
                newState['_world'] = destState
                newState['_epoch'] = newState['_epoch'] + 1
                newState['_prob'] = s['_prob'] \
                                    * self.probability(s['_world'],
                                                       destState,action)
                if s.has_key('_parent'):
                    newState['_parent'] = s
                    s['_children'][action] = newState
                try:
                    del newState['_messages']
                except KeyError:
                    pass
                states.append(newState)
            states.remove(s)
        
    # The state estimators.  These default methods implement "perfect
    # recall"

    # SE^0_i
    def initialStateEstimator(self,agent):
	"""Generates the initial belief state for the specified
	agent"""
	return []

    # SE_i\bullet\Sigma
    def preComStateEstimator(self,agent,beliefs,obs,epoch=-1):
	"""Updates the agent's beliefs based on its observations"""
	newBeliefs = beliefs[:]
	newBeliefs.insert(0,obs)
        if epoch >= 0:
            newBeliefs[0]['_epoch'] = epoch
	return newBeliefs

    # SE_i\Sigma\bullet
    def postComStateEstimator(self,agent,beliefs,msgs,epoch=-1):
	"""Updates the agent's beliefs based on received messages"""
	newBeliefs = beliefs[:]
	newBeliefs.insert(0,msgs)
        if epoch >= 0:
            newBeliefs[0]['_epoch'] = epoch
	return newBeliefs
	
    def updateBeliefs(self,world,beliefs,agent):
	"""Generates all possible new belief states for an agent"""
	
        
##	    # Communication phase
##	    if debug:
##		print
##		print '--------'
##		print 'Communication phase:'
##	    for s in stateList[:]:
##		if debug:
##		    print '--------'
##		    print 'Examining state:',s
##                # Find messages sent in past
##                for msg in beliefs:
##                    if msg['_type'] == 'message' and \
##                       msg['_epoch'] == epoch:
##                        break
##                newState = copy.copy(s)
##                if debug:
##                    print 'Messages:',msg
##                # Update agents' beliefs based on messages exchanged
##                for agent in self.mtdp.agents:
##                    newState[agent.name] = self.mtdp.postComStateEstimator(agent,s[agent.name],msg)
##                # Update value based on communication cost
##                newState['_value'] = self.mtdp.rewardCom(s['_world'],msg)
##                newState['_parent'] = s
##                stateList.append(newState)
##                stateList.remove(s)
